#include <pthread.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/uio.h>
#include <sys/stat.h>
#include <sys/times.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <poll.h>
#include <signal.h>
#include "processor.h"
#include "parse.h"
#include "missing.h"

#define exit(a) do { fflush(stdout); fflush(stderr); _exit(a); } while(0)

#define SOCK "/dev/netctl"

#define TERM	0
#define STATUS	1
#define V4ADDR	2
#define V6ADDR	3

typedef struct ifdata
{
	struct ifdata *next;
	char name[IFNAMSIZ+1];
	int state;
	int v4total;
	int v6total;
	struct in_addr *v4;
	struct in6_addr *v6;
	char data[0];
} IFDATA;

static OPDEV *da=NULL;
static OPCOND *ca=NULL;
static IFDATA *ia=NULL;
static int simulate=0;
static int maxwait=10000;
static int adjust;
static int sfd;
static int ufd;
static char *script;
static pthread_mutex_t ltx=PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t mtx=PTHREAD_MUTEX_INITIALIZER;

static int clisend(int cmd,void *data,int len)
{
	int i;
	struct sockaddr_un a;
	struct iovec iov[2];

	memset(&a,0,sizeof(a));
	a.sun_family=AF_UNIX;
	strcpy(a.sun_path,SOCK);

	iov[0].iov_base=&cmd;
	iov[0].iov_len=sizeof(int);
	iov[1].iov_base=data;
	iov[1].iov_len=len;

	if((i=socket(PF_UNIX,SOCK_STREAM|SOCK_CLOEXEC,0))==-1)
	{
		perror("socket");
		return -1;
	}

	if(connect(i,(struct sockaddr *)(&a),sizeof(a)))
	{
		perror("connect");
		close(i);
		return -11;
	}

	if(writev(i,iov,data?2:1)!=(data?len:0)+sizeof(int))
	{
		perror("writev");
		close(i);
		return -1;
	}

	return i;
}

static int locked_fork(void)
{
	int s;
	int r;

	pthread_mutex_lock(&ltx);
	r=fork();
	s=errno;
	pthread_mutex_unlock(&ltx);
	errno=s;
	return r;
}

static int locked_accept(int fd,struct sockaddr *a,socklen_t *l)
{
	int r;
	int s;

	pthread_mutex_lock(&ltx);
	r=accept(fd,a,l);
	s=errno;
	if(r!=-1)if(fcntl(r,F_SETFD,FD_CLOEXEC))
	{
		s=errno;
		close(r);
		r=-1;
	}
	pthread_mutex_unlock(&ltx);
	errno=s;
	return r;
}

static void evdump(IFDATA *d)
{
	int i;
	char bfr[INET6_ADDRSTRLEN];

	printf("event %s for %s (%d IPv4, %d IPV6)\n",d->state==UP?"UP":"DOWN",
		d->name,d->v4total,d->v6total);

	for(i=0;i<d->v4total;i++)
	{
		inet_ntop(AF_INET,&d->v4[i],bfr,INET6_ADDRSTRLEN);
		printf("\tIPv4: %s\n",bfr);
	}

	for(i=0;i<d->v6total;i++)
	{
		inet_ntop(AF_INET6,&d->v6[i],bfr,INET6_ADDRSTRLEN);
		printf("\tIPv6: %s\n",bfr);
	}
}

static int v4sort(const void *s1,const void *s2)
{
	return memcmp(s1,s2,sizeof(struct in_addr));
}

static int v6sort(const void *s1,const void *s2)
{
	return memcmp(s1,s2,sizeof(struct in6_addr));
}

static IFDATA *dorecv(int queue)
{
	int total;
	IFDATA *data;
	struct iovec iov[5];

	if(read(queue,&total,sizeof(total))!=sizeof(total))
	{
		perror("read");
		return NULL;
	}

	iov[0].iov_len=IFNAMSIZ+1;
	iov[1].iov_len=sizeof(data->state);
	iov[2].iov_len=sizeof(data->v4total);
	iov[3].iov_len=sizeof(data->v6total);
	iov[4].iov_len=total-iov[0].iov_len-iov[1].iov_len-
			iov[2].iov_len-iov[3].iov_len;


	if(!(data=malloc(sizeof(IFDATA)+iov[4].iov_len)))
	{
		perror("malloc");
		return NULL;
	}

	iov[0].iov_base=data->name;
	iov[1].iov_base=&data->state;
	iov[2].iov_base=&data->v4total;
	iov[3].iov_base=&data->v6total;
	iov[4].iov_base=data->data;

	if(readv(queue,iov,5)!=total)
	{
		perror("readv");
		free(data);
		return NULL;
	}

	data->next=NULL;
	data->v4=(struct in_addr *)data->data;
	data->v6=(struct in6_addr *)(data->data+
			data->v4total*sizeof(struct in_addr));

	if(data->v4total>1)qsort(data->v4,data->v4total,sizeof(struct in_addr),
		v4sort);

	if(data->v6total>1)qsort(data->v6,data->v6total,sizeof(struct in6_addr),
		v6sort);

	return data;
}

static void validate(OPSTATE *s,int nvalid,int *changed)
{
	if(s->valid==nvalid)return;

	if(simulate)
	{
		if(nvalid)printf("state %s: FALSE->TRUE\n",s->name);
		else printf("state %s: TRUE->FALSE\n",s->name);
	}

	s->valid=nvalid;
	*changed=1;
}

static void trigger(OPSTATE *s,int nvalid,int *changed)
{
	if(!nvalid)
	{
		s->valid=0;
		return;
	}

	if(simulate)printf("state %s: CHANGE\n",s->name);

	s->valid=1;
	*changed=1;
}

static void evmgr(char *dev,IFDATA *new,IFDATA *old)
{
	int i;
	int j;
	int k;
	OPDEV *d;
	OPSTATE *s;
	OPCOND *c;

	if(new&&old)
	{
		if(new->v4total==old->v4total)
		{
			for(i=0;i<new->v4total;i++)if(memcmp(&new->v4[i],
				&old->v4[i],sizeof(struct in_addr)))break;

			if(i==new->v4total)i=0;
			else i=1;
		}
		else i=1;

		if(new->v6total==old->v6total)
		{
			for(j=0;j<new->v6total;j++)if(memcmp(&new->v6[j],
				&old->v6[j],sizeof(struct in6_addr)))break;

			if(j!=new->v6total)i|=2;
		}
		else i|=2;

		if(!i)return;
	}
	else i=0;

	for(j=0,d=da;d;d=d->next)if(!strcmp(d->dev,dev))
	{
		j=1;

		if(!new)
		{
			for(s=d->states[UP];s;s=s->next)validate(s,0,&j);
			for(s=d->states[DOWN];s;s=s->next)validate(s,1,&j);
			continue;
		}

		for(s=d->states[DOWN];s;s=s->next)validate(s,0,&j);
		for(s=d->states[UP];s;s=s->next)
		{
			if(s->match==NOIP)validate(s,1,&j);
			else if(s->match==IPV4&&new->v4total)validate(s,1,&j);
			else if(s->match==IPV4&&!new->v4total)validate(s,0,&j);
			else if(s->match==IPV6&&new->v6total)validate(s,1,&j);
			else if(s->match==IPV6&&!new->v6total)validate(s,0,&j);
			else if(new->v4total||new->v6total)validate(s,1,&j);
			else validate(s,0,&j);
		}

		if(!old)
		{
			for(s=d->states[CHANGE];s;s=s->next)trigger(s,0,&j);
			continue;
		}

		for(s=d->states[CHANGE];s;s=s->next)
		{
			if(s->match==IPV4)trigger(s,i&1,&j);
			else if(s->match==IPV6)trigger(s,i&2,&j);
			else trigger(s,i&3,&j);
		}


	}
	else for(s=d->states[CHANGE];s;s=s->next)trigger(s,0,&j);

	if(!j)return;

	for(c=ca;c;c=c->next)
	{
		for(i=0,j=1,k=0;i<c->total;i++)switch(c->states[i]->type)
		{
		case UP:
		case DOWN:
			if(!c->states[i]->valid)j=0;
			break;

		case CHANGE:
			if(c->states[i]->valid)k=1;
			break;
		}

		if(j)
		{
			if(c->nextstate==STOPPED)
			{
				c->nextstate=STARTED;
				c->delay=c->start;
			}
		}
		else
		{
			if(c->nextstate==STARTED)
			{
				c->nextstate=STOPPED;
				c->delay=c->stop;
			}
		}

		if(c->currstate==STARTED&&c->nextstate==STARTED&&k)c->event=1;
	}
}

static void runscript(char *p1,char *p2)
{
	pid_t p;
	uint64_t dummy;
	struct pollfd pp;

	pp.fd=sfd;
	pp.events=POLLIN;

	dummy=read(sfd,&dummy,sizeof(dummy));

	switch((p=locked_fork()))
	{
	case 0:	execl(script,script,p1,p2,NULL);
		perror("exec");
		exit(255);

	case -1:perror("fork");
		break;

	default:
		while(poll(&pp,1,maxwait)<=0)kill(p,SIGTERM);
		break;
	}
}

static int condhandler(int expired)
{
	int newex;
	OPCOND *c;
	clock_t t;

repeat:	t=times(NULL);

	for(newex=-1,c=ca;c;c=c->next)if(c->currstate!=c->nextstate)
	{
		if(expired!=-1)
		{
			if(expired>c->delay)c->delay=0;
			else c->delay-=expired;
		}

		if(c->delay)
		{
			if(newex==-1)newex=c->delay;
			else if(newex>c->delay)newex=c->delay;
		}
		else
		{
			if(simulate)printf("exec: %s %s\n",
				c->nextstate==STARTED?"start":"stop",c->label);
			else runscript(c->nextstate==STARTED?"start":"stop",
				c->label);

			c->currstate=c->nextstate;
			if(c->currstate==STOPPED)c->event=0;
		}
	}
	else if(c->currstate==STARTED&&c->event)
	{
		if(simulate)printf("exec: change %s\n",c->label);
		else runscript("change",c->label);

		c->event=0;
	}

	if(newex!=-1)
	{
		t=times(NULL)-t;
		t*=adjust;
		if(t<0)newex=0;
		else
		{
			if(t>newex)t=newex;
			else t*=adjust;
			if(t>newex)newex=0;
			else newex-=t;
		}

		if(!newex)goto repeat;
	}

	return newex;
}

void *prcjob(void *data)
{
	IFDATA *d;
	IFDATA **r;
	struct pollfd p[2];
	sigset_t s;
	int delay=-1;
	clock_t t;

	sigfillset(&s);
	pthread_sigmask(SIG_BLOCK,&s,NULL);

	p[0].fd=((int *)data)[0];
	p[1].fd=((int *)data)[1];
	p[0].events=POLLIN;
	p[1].events=POLLIN;

	if(simulate)printf("exec: init\n");
	else runscript("init",NULL);

	while(1)
	{
		delay=condhandler(delay);

		t=times(NULL);

		switch(poll(p,2,delay))
		{
		case -1:
		case 0:	goto calc;
		}

		if(p[0].revents&POLLIN)break;

		if(p[1].revents&POLLIN)
		{
			if(!(d=dorecv(p[1].fd)))break;

			if(simulate)evdump(d);

			for(r=&ia;*r;r=&(*r)->next)
				if(!strcmp((*r)->name,d->name))break;

			switch(d->state)
			{
			case DOWN:
				evmgr(d->name,NULL,*r);
				free(d);
				if(*r)
				{
					pthread_mutex_lock(&mtx);
					d=*r;
					*r=d->next;
					pthread_mutex_unlock(&mtx);
					free(d);
				}
				break;

			case UP:
				evmgr(d->name,d,*r);
				pthread_mutex_lock(&mtx);
				if(*r)
				{
					d->next=(*r)->next;
					free(*r);
				}
				*r=d;
				pthread_mutex_unlock(&mtx);
				break;
			}
		}

calc:		t=times(NULL)-t;
		if(t<0)delay=100000;
		else
		{
			if(t>100000)t=100000;
			else t*=adjust;
			if(t>100000)delay=100000;
			else delay=t;
		}
	}

	if(simulate)printf("exec: term\n");
	else runscript("term",NULL);

	pthread_exit(NULL);
	return NULL;
}

void *unixjob(void *data)
{
	int c;
	int cmd;
	IFDATA *d;
	struct pollfd p[2];
	struct iovec iov[2];
	char name[IFNAMSIZ+1];
	sigset_t s;

	sigfillset(&s);
	pthread_sigmask(SIG_BLOCK,&s,NULL);

	p[0].fd=(int)((long)data);
	p[1].fd=ufd;
	p[0].events=POLLIN;
	p[1].events=POLLIN;

	while(1)
	{
		if(poll(p,2,-1)<=0)continue;

		if(p[0].revents&POLLIN)break;

		if(!(p[1].revents&POLLIN))continue;

		if((c=locked_accept(ufd,NULL,NULL))==-1)continue;

		if(read(c,&cmd,sizeof(cmd))!=sizeof(cmd))goto done;

		switch(cmd)
		{
		case TERM:
			kill(getpid(),SIGTERM);
			break;

		case STATUS:
		case V4ADDR:
		case V6ADDR:
			if(read(c,name,sizeof(name))!=sizeof(name))
			{
				perror("read");
				break;
			}

			pthread_mutex_lock(&mtx);

			for(d=ia;d;d=d->next)if(!strcmp(d->name,name))break;

			switch(cmd)
			{
			case STATUS:
				if(!d||d->state==DOWN)cmd=0;
				else cmd=1;
				if(write(c,&cmd,sizeof(cmd))!=sizeof(cmd))
					perror("write");
				break;

			case V4ADDR:
				if(d)cmd=d->v4total;
				else cmd=0;
				iov[0].iov_base=&cmd;
				iov[0].iov_len=sizeof(cmd);
				iov[1].iov_base=d?d->v4:NULL;
				iov[1].iov_len=
					d?d->v4total*sizeof(struct in_addr):0;
				if(writev(c,iov,d?2:1)!=
					iov[0].iov_len+iov[1].iov_len)
						perror("writev");
				break;

			case V6ADDR:
				if(d)cmd=d->v6total;
				else cmd=0;
				iov[0].iov_base=&cmd;
				iov[0].iov_len=sizeof(cmd);
				iov[1].iov_base=d?d->v6:NULL;
				iov[1].iov_len=
					d?d->v6total*sizeof(struct in6_addr):0;
				if(writev(c,iov,d?2:1)!=
					iov[0].iov_len+iov[1].iov_len)
						perror("writev");
				break;
			}

			pthread_mutex_unlock(&mtx);
			break;
		}

done:		close(c);
	}

	close(ufd);
	unlink(SOCK);

	pthread_exit(NULL);
	return NULL;
}

int prcprep(char *fn,char *scr,int sigfd,int max,int trace)
{
	int c;
	struct sockaddr_un a;
	mode_t m;

	adjust=1000/sysconf(_SC_CLK_TCK);
	if(!adjust)
	{
		fprintf(stderr,"illegal time resolution\n");
		return -1;
	}

	if(access(scr,X_OK))
	{
		perror("access");
		return -1;
	}
	script=scr;

	if(parse(fn,&da,&ca))return -1;

	memset(&a,0,sizeof(a));
	a.sun_family=AF_UNIX;
	strcpy(a.sun_path,SOCK);

	if(!access(SOCK,W_OK))
	{
		if((c=socket(PF_UNIX,SOCK_STREAM|SOCK_CLOEXEC,0))==-1)
		{
			perror("socket");
			return -1;
		}

		if(!connect(c,(struct sockaddr *)(&a),sizeof(a)))
		{
			close(c);
			fprintf(stderr,"already running\n");
			return -1;
		}

		close(c);
		unlink(SOCK);
	}

	if((ufd=socket(PF_UNIX,SOCK_STREAM|SOCK_CLOEXEC|SOCK_NONBLOCK,0))==-1)
	{
		perror("socket");
		close(ufd);
		return -1;
	}

	m=umask(0);
	if(bind(ufd,(struct sockaddr *)(&a),sizeof(a)))
	{
		perror("bind");
		umask(m);
		close(ufd);
		return -1;
	}
	umask(m);

	if(listen(ufd,5))
	{
		perror("listen");
		close(ufd);
		unlink(SOCK);
		return -1;
	}

	sfd=sigfd;
	maxwait=max*1000;
	simulate=trace;

	return 0;
}

int termreq(void)
{
	int c;

	if((c=clisend(TERM,NULL,0))==-1)return -1;
	close(c);
	return 0;
}

int statreq(char *ifc)
{
	int c;
	int status;
	char name[IFNAMSIZ+1];

	strncpy(name,ifc,IFNAMSIZ);
	name[IFNAMSIZ]=0;

	if((c=clisend(STATUS,name,sizeof(name)))==-1)return -1;

	if(read(c,&status,sizeof(status))!=sizeof(status))
	{
		perror("read");
		close(c);
		return -1;
	}

	close(c);

	if(status)printf("up\n");
	else printf("down\n");

	return 0;
}

int v4req(char *ifc)
{
	int c;
	int total;
	struct in_addr *a;
	char name[IFNAMSIZ+1];
	char bfr[INET_ADDRSTRLEN];

	strncpy(name,ifc,IFNAMSIZ);
	name[IFNAMSIZ]=0;

	if((c=clisend(V4ADDR,name,sizeof(name)))==-1)return -1;

	if(read(c,&total,sizeof(total))!=sizeof(total))
	{
		perror("read");
		close(c);
		return -1;
	}

	if(!total)
	{
		close(c);
		return 0;
	}

	if(!(a=malloc(total*sizeof(struct in_addr))))
	{
		perror("malloc");
		close(c);
		return -1;
	}

	if(read(c,a,total*sizeof(struct in_addr))!=total*sizeof(struct in_addr))
	{
		perror("read");
		close(c);
		return -1;
	}

	close(c);

	for(c=0;c<total;c++)
	{
		inet_ntop(AF_INET,&a[c],bfr,INET_ADDRSTRLEN);
		printf("%s\n",bfr);
	}

	free(a);

	return 0;
}

int v6req(char *ifc)
{
	int c;
	int total;
	struct in6_addr *a;
	char name[IFNAMSIZ+1];
	char bfr[INET6_ADDRSTRLEN];

	strncpy(name,ifc,IFNAMSIZ);
	name[IFNAMSIZ]=0;

	if((c=clisend(V6ADDR,name,sizeof(name)))==-1)return -1;

	if(read(c,&total,sizeof(total))!=sizeof(total))
	{
		perror("read");
		close(c);
		return -1;
	}

	if(!total)
	{
		close(c);
		return 0;
	}

	if(!(a=malloc(total*sizeof(struct in6_addr))))
	{
		perror("malloc");
		close(c);
		return -1;
	}

	if(read(c,a,total*sizeof(struct in6_addr))!=
			total*sizeof(struct in6_addr))
	{
		perror("read");
		close(c);
		return -1;
	}

	close(c);

	for(c=0;c<total;c++)
	{
		inet_ntop(AF_INET6,&a[c],bfr,INET6_ADDRSTRLEN);
		printf("%s\n",bfr);
	}

	free(a);

	return 0;
}
